unit POIFS;

{
  POIFS for Delphi 32
  Copyright (C) 2004 Paul Toth <tothpaul@free.fr>
  http://tothpaul.free.fr

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

}

// http://jakarta.apache.org/poi/poifs/fileformat.html

{

NB: you can also use Windows API :D

Uses
 ActiveX, ComObj;
var
 STG:IStorage;
 DOC:IStream;
 XL8:TMemoryStream; // a BIFF8 stream
 Len:cardinal;
begin
 OleCheck(StgCreateDocFile('Classeur1.xls',STGM_DIRECT or STGM_SHARE_EXCLUSIVE or STGM_WRITE,0,Stg));
 OleCheck(Stg.CreateStream('Workbook',STGM_DIRECT or STGM_SHARE_EXCLUSIVE or STGM_WRITE,0,0,Doc));
 OleCheck(Doc.Write(XL8.Memory,XL8.Size,@Len));
end;

}

interface

uses
 Classes,SysUtils,Dialogs;

const
 Magic=$E11AB1A1E011CFD0;

type
 TOLEHeader=packed record // 512 bytes
  Sign     : int64;    // $E11AB1A1E011CFD0 (magic)
  UK1      : integer;  // 0
  UK2      : integer;  // 0
  UK21     : integer;  // 0
  UK3      : integer;  // 0
  UK4      : word;     // $003E
  UK5      : word;     // $0003
  UK6      : Smallint; // -2
  BigSize  : word;     // 9 (2^9=512 bytes)
  SmallSize: integer;  // 6 (2^6= 64 bytes)
  UK7      : integer;  // 0
  UK8      : integer;  // 0
  BATCount : integer;  // 1
  DIRIndex : integer;  // 25
  UK9      : integer;  // 0
  UK10     : integer;  // $1000
  SBATIndex: integer;  // -2
  SBATCount: integer;  //  0
  XBATIndex: integer;  // -2
  XBATCount: integer;  //  0
  BATArray : array[0..108] of integer; // 24,-1,-1,-1...
 end;

 TDIREntry=packed record  // 128 bytes (4 entries by Block)
  Name    : array[0..31] of WideChar;
  NameLen : word;
  Kind    : byte; // 1=Dir, 2=File, 5=RootEntry
  Color   : byte; // 0=Red, 1=Black
  Prior   : integer;
  Next    : integer;
  Child   : integer; // first child
  dummy1  : array[$50..$73] of byte; // Writer CLSID ?!
  Start   : integer;
  Size    : integer;
  dummy2  : array[$7C..$7F] of byte; // Date ?!
 end;

 TPOIFS=class
 private
  fStream    :TFileStream; // OLE2 Storage
  fHeader    :TOLEHeader;  // File Header
  fBATIndex  :integer;     // Current BAT array Index
  fBATCache  :array[0..127] of integer; // Current BAT block
  fBATChange :boolean;    // current BAT block need to be saved
  fDIRIndex  :integer;     // Current DIR block index
  fDIRCache  :array[0..3] of TDIREntry; // current DIR block
  fDIRChange :boolean;    // current DIR block need to be saved
  fSBATIndex :integer;
  fSBATBlock :integer;
  fSBATStart :integer; // from RootEntry
  fSBATCache :array[0..127] of integer;
  fSBATChange:boolean;
  fSBLKCount :integer;
  fSBLKIndex :integer;
  fSBLKBlock :integer;
  fSBLKCache :array[0..511] of byte;
  fSBLKChange:boolean;
  procedure ClearDIRCache; // create an empty DIR block
  function NextBlock(Index:integer; Extend:boolean=False):integer;
  function NextBlockIsNew(var Index:integer):boolean;
  function ExtendBlock(Index:integer):integer;
  function AllocBlocks(Count:integer):integer; // alloc some data blocks
  function AllocSmallBlocks(Count:integer):integer;
  function AllocEntry:integer; // alloc a new entry in the DIR blocks
  procedure AddEntry(FileName:string; FileSize,FirstBlock:integer); // add a file entry
  function DirEntry(Index:integer):integer; // return DIRCache Block index for a given DIR index
  procedure SaveBlock(Index:integer; var Block);
  procedure ReadBlock(Index:integer; var Block);
  procedure ReadSmallBlock(Index:integer; var Block);
  procedure SaveSmallBlock(Index:integer; var Block);
  procedure SetBATIndex(Value:integer); // read (and save if needed) fBATCache
  procedure SetDIRIndex(Value:integer); // read (and save if needed) fDIRCache
  procedure SetSBATIndex(Value:integer); // read (and save if needed) fBATCache
  procedure SetSBLKIndex(Value:integer);
  procedure SaveHeader;
  procedure SaveBAT;
  procedure SaveDIR;
  procedure SaveSBAT;
  procedure SaveSBLK;
  property BATIndex:integer read fBATIndex write SetBATIndex;
  property DIRIndex:integer read fDIRIndex write SetDIRIndex;
  property SBATIndex:integer read fSBATIndex write SetSBATIndex;
  property SBLKIndex:integer read fSBLKIndex write SetSBLKIndex;
 public
  destructor Destroy; override;
  procedure OpenFile(AFileName:string); // open an existing file
  procedure CreateFile(AFileName:string); // create a new file
  procedure AddStream(FileName:string; Stream:TStream; AllowSmallBlocks:boolean=False);
  function GetStream(FileName:string; Stream:TStream):boolean;
  procedure Close;
 end;

implementation

{ TPOIFS }

resourcestring
 SNotPOIFS='Ce fichier n''est pas au format OLE2';
 SUnusualFileFormat='Format de fichier inhabituel';
 SUnsupportedXBAT='Format de fichier non support (XBAT)';
 SFileNotOpen='Aucun fichier ouvert';

destructor TPOIFS.Destroy;
begin
 Close;
 inherited;
end;

procedure TPOIFS.OpenFile(AFileName:string);
begin
// release previous file
 Close;
// open new one
 fStream:=TFileStream.Create(AFileName,fmOpenRead);
 try
 // check header
  fStream.ReadBuffer(fHeader,SizeOf(fHeader));
  if fHeader.Sign<>Magic then raise Exception.Create(SNotPOIFS);
  if (fHeader.UK1<>0)
  or (fHeader.UK2<>0)
  or (fHeader.UK21<>0)
  or (fHeader.UK3<>0)
  or (fHeader.UK4<>$3E)or(fHeader.UK5<>3)or(fHeader.UK6<>-2)
  or (fHeader.BigSize<>9)or(fHeader.SmallSize<>6)
  or (fHeader.UK7<>0)or(fHeader.UK8<>0)or(fHeader.UK9<>0)or(fHeader.UK10<>$1000) then raise Exception.Create(SUnusualFileFormat);
// todo: support of Extra BAT blocks (for big files > 109*128*512 bytes)
  if (fHeader.XBATCount>0)or(fHeader.XBATIndex<>-2) then raise Exception.Create(SUnsupportedXBAT);
  if fHeader.SBATCount>0 then begin
   DIRIndex:=fHeader.DIRIndex;
   fSBATStart:=fDIRCache[0].Start;
   fSBLKCount:=fDIRCache[0].Size div 64;
  end;
 except
  FreeAndNil(fStream);
  raise;
 end;
end;

procedure TPOIFS.CreateFile(AFileName:string);
var
 ws:WideString;
begin
 Close;
 FreeAndNil(fStream);
 FillChar(fHeader,SizeOf(fHeader),0);
 FillChar(fHeader.BATArray,SizeOf(fHeader.BATArray),255); // -1
 fHeader.Sign:=Magic;
 fHeader.UK4:=$3E; fHeader.UK5:=3; fHeader.UK6:=-2;
 fHeader.BigSize:=9; fHeader.SmallSize:=6;
 fHeader.UK10:=$1000;
 fHeader.SBATIndex:=-2;
 fHeader.XBATIndex:=-2;
 fHeader.BATCount:=1;
 fHeader.DIRIndex:=1;    // Block 1 is the root entry
 fHeader.BATArray[0]:=0; // a single BAT block at offset 0
 Fillchar(fBATCache,SizeOf(fBATCache),255); // -1
 fBATCache[0]:=-3; // BATCache himself
 fBATCache[1]:=-2; // DIRCache
 // NB: Excel 2000 place this blocks after the data blocks...don't care :D
 ClearDIRCache;
 with fDIRCache[0] do begin
  ws:='Root Entry';
  move(ws[1],Name,2*Length(ws));
  NameLen:=2*Length(WS)+2;
  Kind := 5;
  Color:= 1;
  Start:=-2;
  (*
  dummy1[$50]:=32;
  dummy1[$51]:=8;
  dummy1[$52]:=2;
  dummy1[$58]:=192;
  dummy1[$5F]:=70;
  *)
 end;
 fStream:=TFileStream.Create(AFileName,fmCreate);
 try
  fStream.WriteBuffer(fHeader,SizeOf(fHeader));
  fStream.WriteBuffer(fBATCache,SizeOf(fBATCache));
  fStream.WriteBuffer(fDIRCache,SizeOf(fDIRCache));
  fDIRIndex:=1;
  fBATIndex:=0;
 except
  FreeAndNil(fStream);
  raise;
 end;
end;

// alloc a DIR entry
function TPOIFS.AllocEntry:integer;
var
 di:integer;
  i:integer;
begin
// first entry
 Result:=0;
// first DIR block
 di:=fHeader.DIRIndex; // Load RootEntry
 repeat
 // read the DIR block (fDIRCache)
  DIRIndex:=di;
  // look for an empty entry
  for i:=0 to 3 do begin
   with fDIRCache[i] do begin
   // do I need to check all this ?!
    if (Kind=0)and(Next=-1)and(Prior=-1)and(Child=-1) then exit; // Found an empty Entry
   end;
   inc(Result); // Next entry (absolute index)
  end;
  // find the next DIR block in the BAT array
  if NextBlockIsNew(di) then ClearDirCache;
  (*
  BATIndex:=di div 128;
  if fBATCache[di mod 128]=-2 then begin // Last block
   xi:=AllocBlocks(1); // Alloc a new block
   BATIndex:=di div 128; // previous block
   fBATCache[di mod 128]:=xi; // chain blocks
   fBATChange:=True;
   ClearDirCache; // create empty dir entries
   di:=xi; // select this block
  end else begin
   di:=fBATCache[di mod 128]; // select the new block
  end;
  *)
 until False; // forever...
end;

// add a new DIR entry
procedure TPOIFS.AddEntry(FileName:string; FileSize,FirstBlock:integer);
var
 e:integer;
 ws:WideString;
 i,n:integer;
begin
// alloc a new entry
 e:=AllocEntry;
// setup
 with fDIRCache[DirEntry(e)] do begin
  ws:=FileName;
  Move(ws[1],Name,2*Length(FileName));
  NameLen:=2+2*Length(FileName);
  Kind :=2;
  Color:=1;
  Start:=Firstblock;
  Size :=FileSize;
 end;
 fDIRChange:=True;
// root entry
 DirEntry(0);
 i:=fDIRCache[0].Child; // RootEntry.Child
 if i<0 then begin
 // First file
  fDIRCache[0].Child:=e;
  fDIRChange:=True;
 end else begin
 // todo: use Next, Prior & Color to sort Streams in the file
  n:=fDIRCache[DirEntry(i)].Next;
  while n<>-1 do begin
   i:=n;
   n:=fDIRCache[DirEntry(i)].Next;
  end;
  fDIRCache[DirEntry(i)].Next:=e;
  fDIRChange:=True;
 end;
 SaveDIR;
end;

// read fDIRCache and return fDIRCache index for a given DIR absolute index
function TPOIFS.DirEntry(Index:integer):integer;
var
 i:integer;
 b:integer;
begin
 i:=Index div 4; // 4 entries per Block
 b:=fHeader.DIRIndex; // First block
 while i>0 do begin
  BATIndex:=b div 128; // 128 index per BAT block
  b:=fBATCache[b]; // Next Block
  dec(i);
 end;
 DIRIndex:=b; // read fDIRCache
 Result:=Index mod 4; // index in fDIRCache
end;

procedure TPOIFS.SaveBlock(Index:integer; var Block);
begin
 fStream.Position:=512+Index*512;
 fStream.WriteBuffer(Block,512);
end;

procedure TPOIFS.ReadBlock(Index:integer; var Block);
begin
 fStream.Position:=512+Index*512;
 if fStream.Position>=fStream.Size then
  FillChar(Block,512,0)
 else
  fStream.ReadBuffer(Block,512);
end;

procedure TPOIFS.ReadSmallBlock(Index:integer; var Block);
begin
 SBLKIndex:=Index;
 Move(fSBLKCache[64*(Index mod 8)],Block,64);
end;

procedure TPOIFS.SaveSmallBlock(Index:integer; var Block);
begin
 if fSBATStart<0 then begin
  fSBATStart:=AllocBlocks(1);
  DIRIndex:=fHeader.DIRIndex;
  (*
  FillChar(fDIRCache[0].dummy1,SizeOf(fDIRCache[0].dummy1),0);
  fDIRCache[0].dummy1[$73]:=1;
  fDIRCache[0].dummy1[$72]:=195;
  fDIRCache[0].dummy1[$71]:=236;
  fDIRCache[0].dummy1[$70]:=185;
  fDIRCache[0].dummy1[$6F]:=46;
  fDIRCache[0].dummy1[$6E]:=246;
  fDIRCache[0].dummy1[$6D]:=60;
  fDIRCache[0].dummy1[$6C]:=96;
  *)
  fDIRCache[0].Start:=fSBATStart;
  // fDIRCache[0].Size:=64*fSBLKCount;
  fDIRChange:=True;
  SaveDIR;
 end;
{ SBLKCache contient 8 SmallBlock de 64 octets }
 SBLKIndex:=Index;
 Move(Block,fSBLKCache[64*(Index mod 8)],64);
 fSBLKChange:=True;
end;

procedure TPOIFS.SetBATIndex(Value:integer);
begin
 if fBATIndex=Value then exit;
 if fBATIndex>=fHeader.BATCount then raise Exception.Create('BATIndex overflow');
 SaveBAT;
 fBATIndex:=Value;
 ReadBlock(fHeader.BATArray[fBATIndex],fBATCache);
end;

procedure TPOIFS.SetDIRIndex(Value:integer);
begin
 if fDIRIndex=Value then exit;
 SaveDIR;
 fDIRIndex:=Value;
 ReadBlock(fDIRIndex,fDIRCache);
end;

procedure TPOIFS.SetSBATIndex(Value: integer);
begin
 if fSBATIndex=Value then exit;
 SaveSBAT;
 fSBATIndex:=Value;
 fSBATBlock:=fHeader.SBATIndex;
 while Value>128 do begin
  BATIndex:=fSBATBlock div 128;
  fSBATBlock:=fBATCache[fSBATBlock mod 128];
  dec(Value,128);
 end;
 ReadBlock(fSBATBlock,fSBATCache);
end;

procedure TPOIFS.SetSBLKIndex(Value:integer);
var
 SameBlock:boolean;
 NextBlock:integer;
begin
// on ne change pas de small block
 if fSBLKIndex=Value then exit;
// change-t-on de block (8 small block par block)
 SameBlock:=(fSBLKIndex>=0)and((fSBLKIndex div 8)=(Value div 8));
 fSBLKIndex:=Value;
 if SameBlock then exit;
// enregistrer le block prcdent
 SaveSBLK;
// on part du premier groupe de 8 smallblock
 fSBLKBlock:=fSBATStart;
// tant qu'on le dpasse
 while Value>=8 do begin
 // checher dans la BATArray le block qui suit
  if NextBlockIsNew(fSBLKBlock) then begin
   inc(fHeader.SBATCount);
   SaveHeader;
  end;
 (*
  BATIndex:=fSBLKBlock div 128;
  // extend SBATArray
  NextBlock:=fBATCache[fSBLKBlock mod 128];
  if NextBlock=-2 then begin // Last Block
   NextBlock:=AllocBlocks(1);
   BATIndex:=fSBLKBlock div 128;
   fBATCache[fSBLKBlock mod 128]:=NextBlock;
   fBATChange:=True;
   inc(fHeader.SBATCount);
   SaveHeader;
  end;
  fSBLKBlock:=NextBlock;
 *)
  dec(Value,8);
 end;
 ReadBlock(fSBLKBlock,fSBLKCache);
end;

procedure TPOIFS.SaveHeader;
begin
 fStream.Position:=0;
 fStream.WriteBuffer(fHeader,SizeOf(fHeader));
end;

procedure TPOIFS.SaveBAT;
begin
 if fBATChange then begin
  SaveBlock(fHeader.BATArray[fBATIndex],fBATCache);
  fBATChange:=False;
 end;
end;

procedure TPOIFS.SaveDIR;
begin
 if fDIRChange then begin
  SaveBlock(fDIRIndex,fDIRCache);
  fDIRChange:=False;
 end;
end;

procedure TPOIFS.SaveSBAT;
begin
 if fSBATChange then begin
  SaveBlock(fSBATBlock,fSBATCache);
  fSBATChange:=False;
 end;
end;

procedure TPOIFS.SaveSBLK;
begin
 if fSBLKChange then begin
  SaveBlock(fSBLKBlock,fSBLKCache);
  fSBLKChange:=False;
 end;
end;

procedure TPOIFS.ClearDIRCache;
var
 i:integer;
begin
 FillChar(fDIRCache,SizeOf(fDIRCache),0);
 for i:=0 to 3 do with fDIRCache[i] do begin
  Prior:=-1;
  Next :=-1;
  Child:=-1;
 end;
end;

function TPOIFS.NextBlock(Index:integer; Extend:boolean=False):integer;
begin
 BATIndex:=Index div 128;
 Result:=fBATCache[Index mod 128];
 if Extend and (Result=-2) then Result:=ExtendBlock(Index);
end;

function TPOIFS.NextBlockIsNew(Var Index:integer):boolean;
var
 next:integer;
begin
 BATIndex:=Index div 128;
 next:=fBATCache[Index mod 128];
 Result:=(next=-2);
 if Result then Index:=ExtendBlock(Index) else Index:=next;
end;

function TPOIFS.ExtendBlock(Index:integer):integer;
begin
 Result:=AllocBlocks(1);
 BATIndex:=Index div 128;
 fBATCache[Index mod 128]:=Result;
 fBATChange:=True;
end;

function TPOIFS.AllocBlocks(Count:integer):integer;
var
 i,j,b,l:integer;
begin
 if fHeader.BATCount=0 then raise Exception.Create('BAT count=0 !');
 Result:=-1;
 i:=0;
 while i<=fHeader.BATCount-1 do begin
  BATIndex:=i;
  for j:=0 to 127 do begin
   if fBATCache[j]=-1 then begin
    b:=128*i+j; // block number
    fBATCache[j]:=-2; // this is the last one (for now)
    fBATChange:=True;
    if Result<0 then begin // First block
     Result:=b;
    end else begin
     BATIndex:=l div 128;
     fBATCache[l mod 128]:=b; // chain blocks
     fBATChange:=True;
     BATIndex:=i;
    end;
    dec(Count);
    if Count=0 then begin SaveBAT; exit; end;
    l:=b;
   end;
  end;
  inc(i);
 // Extend file
  if i=fHeader.BATCount then begin
   inc(fHeader.BATCount);
   fHeader.BATArray[i]:=i*128;
   SaveHeader;
   SaveBAT;
   fBATIndex:=i;
   FillChar(fBATCache,SizeOf(fBATCache),255); // -1
   fBATCache[0]:=-3; // BAT himself
   fBATChange:=True;
  end;
 end;
end;

function TPOIFS.AllocSmallBlocks(Count:integer):integer;
var
 i,j,b,l:integer;
begin
// update Root Entry
 inc(fSBLKCount,Count);
 DIRIndex:=fHeader.DIRIndex;
 fDIRCache[0].Size:=64*fSBLKCount;
 fDIRChange:=True;
 // allocate a SBAT array
 if fHeader.SBATIndex=-2 then begin
  fSBATIndex:=0;
  fSBATBlock:=AllocBlocks(1);
  FillChar(fSBATCache,SizeOf(fSBATCache),255); // -1
  fSBATChange:=True;
  fHeader.SBATIndex:=fSBATBlock;
  fHeader.SBATCount:=1;
  SaveHeader;
 end;
 Result:=-1;
 i:=0;
 while i<=fHeader.SBATCount-1 do begin
  SBATIndex:=i;
  for j:=0 to 127 do begin
   if fSBATCache[j]=-1 then begin
    b:=128*i+j; // sblock number
    fSBATCache[j]:=-2; // this is the last one (for now)
    fSBATChange:=True;
    if Result<0 then begin // First block
     Result:=b;
    end else begin
     SBATIndex:=l div 128;
     fSBATCache[l mod 128]:=b; // chain blocks
     fSBATChange:=True;
     SBATIndex:=i;
    end;
    dec(Count);
    if Count=0 then begin SaveSBAT; exit; end;
    l:=b;
   end;
  end;
  inc(i);
 // Extend file
  if i=fHeader.SBATCount then begin
  // todo: test SBAT extension (need more then one small file !):D
   ShowMessage('WARNING ! SBAT extension have not been tested !!!');
   fHeader.SBATCount:=(i+7) div 8;
   SaveHeader;
   SaveSBAT;
   fSBATIndex:=i;
   BATIndex:=fSBATBlock div 128;
   fBATCache[fSBATBlock mod 128]:=AllocBlocks(1);
   fSBATChange:=True;
   fSBATBlock:=fBATCache[fSBATBlock mod 128];
   FillChar(fSBATCache,SizeOf(fSBATCache),255); // -1
   fSBATChange:=True;
  end;
 end;
end;

procedure TPOIFS.AddStream(FileName:string; Stream:TStream; AllowSmallBlocks:boolean=False);
var
 fSize:integer;
 bSize:integer;
 Block:integer;
 Entry:integer;
 Buff :array[0..511] of byte;
begin
 fSize:=Stream.Size;
 Stream.Position:=0;
 if (fSize<4096) and (AllowSmallBlocks=False) then bSize:=4096 else bSize:=fSize;
 if bSize<4096 then begin
  Block:=AllocSmallBlocks((fSize+63) div 64);
  AddEntry(FileName,fSize,Block);
  while fSize>0 do begin
   if Block<0 then raise Exception.Create('SBlock <0 !');
   if fSize>=64 then bSize:=64 else begin
    FillChar(Buff,64,0);
    bSize:=fSize;
   end;
   Stream.ReadBuffer(Buff,bSize);
   SaveSmallBlock(Block,Buff);
   SBATIndex:=Block div 128;
   Block:=fSBATCache[Block mod 128];
   dec(fSize,bSize);
  end;
 end else begin
  Block:=AllocBlocks((bSize+511) div 512);
  AddEntry(FileName,bSize,Block);
  while Block<>-2 do begin
   if fSize>=512 then bSize:=512 else begin
    if bSize>0 then FillChar(Buff,SizeOf(Buff),0);
    bSize:=fSize;
   end;
   if bSize>0 then Stream.ReadBuffer(Buff,bSize);
   SaveBlock(Block,Buff);
   BATIndex:=Block div 128;
   Block:=fBATCache[Block mod 128];
   dec(fSize,bSize);
  end;
 end;
end;

function TPOIFS.GetStream(FileName:string; Stream:TStream):boolean;

 function Browse(Index:integer):integer;
 var
  s:string;
 begin
  with fDIRCache[DirEntry(Index)] do begin
   s:=Name;
   if s=FileName then begin
    Result:=Index;
   end else begin
    if Next>=0 then Result:=Browse(Next) else Result:=-1;
    if (Result<0)and(Prior>=0) then Result:=Browse(Prior);
   end;
  end;
 end;

var
 Index:integer;
 fSize:integer;
 bSize:integer;
 Block:integer;
 Buff :array[0..511] of byte;
begin
 Result:=False;
 Index:=Browse(fDIRCache[DirEntry(0)].Child);
// Index:=DirEntry(5);
 if Index<0 then exit;
 fSize:=fDIRCache[Index].Size;
 Block:=fDIRCache[Index].Start;
 if fSize<4096 then begin // SBAT
  While fSize>0 do begin
   if Block=-2 then raise Exception.Create('End of block');
   ReadSmallBlock(Block,Buff);
   if fSize>=64 then bSize:=64 else bSize:=fSize;
   Stream.WriteBuffer(Buff,bSize);
   dec(fSize,bSize);
   SBATIndex:=Block div 128;
   Block:=fSBATCache[Block mod 128];
  end;
 end else begin // BAT
  While fSize>0 do begin
   if Block=-2 then raise Exception.Create('End of block');
   BATIndex:=Block div 128;
   ReadBlock(Block,Buff);
   if fSize>=512 then bSize:=512 else bSize:=fSize;
   Stream.WriteBuffer(Buff,bSize);
   dec(fSize,bSize);
   Block:=fBATCache[Block mod 128];
  end;
 end;
 Result:=True;
end;

procedure TPOIFS.Close;
begin
 SaveDIR;
 SaveBAT;
 SaveSBAT;
 SaveSBLK;
 fDIRIndex:=-1;
 fBATIndex:=-1;
 fSBATIndex:=-1;
 fSBLKIndex:=-1;
 fSBATStart:=-1;
 fSBLKCount:=0;
 FreeAndNil(fStream);
end;

end.
