unit XL8;

{
  BIFF8 Stream handler for Delphi 32
  Copyright (C) 2004 Paul Toth <tothpaul@free.fr>
  http://tothpaul.free.fr

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

}

// http://sc.openoffice.org/excelfileformat.pdf

{

 some bugs left...

}


interface

uses
 Classes,Graphics;

type
 TObjectList=class(TList)
 public
  procedure Clear; override;
 end;

 TBIFF8Stream=class(TMemoryStream)
 private
  procedure AddChunk(Tag,Size:word);
  procedure VarChunk(Tag:word; const Data; Size:word);
 end;

 TWorkBook=class;

 TWorkSheet=class(TBIFF8Stream)
 private
  fWorkBook:TWorkBook;
  procedure Close;
  function FloatToRK(Float:double; var RK:cardinal):boolean;
  function IntToRK(Int:integer; var RK:cardinal):boolean;
  procedure AddRK(x,y:integer; Value:cardinal);
  procedure AddMULRK(x,y:integer; const RKS:array of cardinal);
 public
  constructor Create;
  procedure AddLabel(x,y:integer; Value:string);
  procedure AddNumber(x,y:integer; Value:double);
  procedure AddFloat(x,y:integer; Value:double);
  procedure AddFloats(x,y:integer; Values:array of double);
  procedure AddInteger(x,y:integer; Value:integer);
  procedure AddIntegers(x,y:integer; Values:array of integer);
 end;

 TCellAlignment=(caGeneral,caLeft,caCentred,caRight,caFilled,caJustified,ceSelectionCentred);
 TCellBorder=(
  cbNone,cbThin,cbMedium,cbDashed,cbDotted,cbThick,cbDouble,cbHair,
  cbMediumDashed,cbThinDashDotted,cbMediumDashDotted,cbThinDashDotDotted,
  cbMediumDashDotDotted,cbSlantedMediumDashDotted
 );

 TFontEntry=class
  Name :string;
  Size :integer;
  Style:TFontStyles;
 end;

 TFontList=class(TObjectList)
 private
  fWorkBook:TWorkBook;
  function FindFont:integer;
  procedure Save;
 public
  constructor Create(WorkBook:TWorkBook);
 end;

 TExtendFormat=class
  Used     :boolean;
  Font     :integer;
  Parent   :integer;
  Flags    :integer;
  Alignment:TCellAlignment;
  Borders  :cardinal;
 end;

 TXFList=class(TObjectList)
 private
  fWorkBook:TWorkBook;
  function FindFormat:integer;
  procedure Save;
 public
  constructor Create(WorkBook:TWorkBook);
 end;

 TWorkbook=class(TBIFF8Stream)
 private
  fWorkSheets:TStringList; // of TWorkSheet
  fSST:TStringList;
  fSSTCount:cardinal;
  fXFList:TXFList;
  fFont     :TFont;
  fAlignment:TCellAlignment;
  fFontList :TFontList;
  fFontNumber:integer;
  fCurrentXF:integer;
  fBorders  :cardinal;
  function AddSST(s:string):integer;
  procedure SetFont(Value:TFont);
  procedure SetAlignment(Value:TCellAlignment);
  function GetBorder(Index:integer):TCellBorder;
  procedure SetBorder(Index:integer; Value:TCellBorder);
  function GetCurrentXF:integer;
 public
  constructor Create;
  destructor Destroy; override;
  procedure BeginDoc;
  function AddWorkSheet(Name:string):TWorkSheet;
  procedure EndDoc;
  property Alignment:TCellAlignment read fAlignment write SetAlignment;
  property Format:integer read GetCurrentXF write fCurrentXF;
  property LeftBorder:TCellBorder index 0 read GetBorder write SetBorder;
  property RightBorder:TCellBorder index 0 read GetBorder write SetBorder;
  property TopBorder:TCellBorder index 0 read GetBorder write SetBorder;
  property BottomBorder:TCellBorder index 0 read GetBorder write SetBorder;
  property Font:TFont read fFont write SetFont;
 end;

const
// BOF
 BIFF8_VERSION   = $0600;
 BIFF8_WB_GLOBAL = $0005;
 BIFF8_VB_MODULE = $0006;
 BIFF8_WORKSHEET = $0010;
 BIFF8_CHART     = $0020;
 BIFF8_MACRO     = $0040;
 BIFF8_WORKSPACE = $0100;

// CHUNKS
 BIFF8_EOF        = $000A;
 BIFF8_FONT       = $0031;
 BIFF8_WINDOW1    = $003D;
 BIFF8_OWNER      = $005C;
 BIFF8_BOUNDSHEET = $0085;
 BIFF8_MULRK      = $00BD;
 BIFF8_XF         = $00E0;
 BIFF8_SST        = $00FC;
 BIFF8_LABELSST   = $00FD;
 BIFF8_EXTSST     = $00FF;
 BIFF8_NUMBER     = $0203;
 BIFF8_LABEL      = $0204; // deprecated -> LABELSST
 BIFF8_RK         = $027E;
 BIFF8_STYLE      = $0293;
 BIFF8_FORMAT     = $041E;
 BIFF8_BOF        = $0809;

type
 TBIFF8_BOF=packed record
  Version     : word;
  DocType     : word;
  BuildID     : word;
  BuildYear   : word;
  HistoryFlag : cardinal;
  LowerVersion: cardinal; // Lowest Excel version for this file
 end;

 TBIFF8_BOUNDSHEET=packed record
  Offset     : cardinal;
  Visibility : byte; // 0..2
  SheetType  : byte; // 0=Worksheet, 2=Chart, 6=VisualBasic module
  NameLen    : byte;
  NameFlag   : byte;
 // SheetName
 end;

 TBIFF8_WINDOW1=packed record
  HorzPos  : word; // Twips
  VertPos  : word;
  Width    : word;
  Height   : word;
  Flags    : word; // 1=Hidden, 2=Minimized, 8=HorzSB, 16=VertSB, 32=ShowTab
  ActiveWS : word;
  ActiveTab: word;
  SelectTab: word;
  TabWidth : word; // in 1/1000 of Width
 end;

 TBIFF8_FONT=packed record
  Height   : word; // Twips
  Flags    : word; // 1=Bold, 2=Italic, 4=Underlined, 8=Struck out
  Color    : word; // Palette or predefined
  Weight   : word; // 400 for normal text, 700 for bold
  Escape   : word; // 0=none, 1=Superscript, 2=Subscript
  Underline: byte; // 0=none, 1=single, 2=double
  Family   : byte; // 0=none, 1=Roman, 2=Swiss, 3=Moderne, 4=Script, 5=Decorative
  Charset  : word; // 0=ANSI Latin, 2=Symbol, 77=Apple Roman, 128=Japanese..., 255=OEM Latin I
  NameLen  : byte;
  NameFlag : byte; // 0=AnsiChar, 1=WideChar
// Name
 end;

 TBIFF8_XF=packed record
  Font   : word; // -> BIFF8_FONT
  Format : word; // -> BIFF8_FORMAT
  Parent : word; // -> BIFF8_XF
  Align  : byte;
  Rotate : byte;
  Indent : byte;
  Flags  : byte;
  Attr1  : cardinal;
  Attr2  : cardinal;
  Attr3  : word;
 end;

 TBIFF8_STYLE=packed record
  XF     : word; // $8000 for Built-in Style
  Style  : byte;
  Level  : byte;
 end;

 TBIFF8_SST=packed record
  Total:cardinal;
  Count:cardinal;
 // SST
 end;

 TBIFF8_EXTSST=packed record
  Count  : word;
  Strings:array[0..7] of record
   StreamOffset : cardinal;
   RecordOffset : word;      // including header (BIFF8_SST & RecorSize)
   NotUsed      : word;
  end;
 end;

 TBIFF8_LABELSST=packed record
  Row : word;
  Col : word;
  XF  : word;
  SST : integer;
 end;

 TBIFF8_LABEL=packed record // deprecated -> LABELSST
  Row : word;
  Col : word;
  XF  : word;
  Len : word;
  Flag: byte; // 0 AnsiChar 1 WideChar
 end;

 TBIFF8_NUMBER=packed record
  Row  : word;
  Col  : word;
  XF   : word;
  Value: double;
 end;

 TBIFF8_RK=packed record
  Row  : word;
  Col  : word;
  XF   : word;
  Value: cardinal; // encoded float or integer
 end;

 TBIFF8_MULRK=packed record
  Row  : word;
  Col  : word;
  // array of XF,RK
  // LastCol
 end;

implementation

//----------------------------------------------------------------------------//

procedure TObjectList.Clear;
var
 i:integer;
begin
 for i:=Count-1 downto 0 do TObject(List[i]).Free;
 inherited;
end;

//----------------------------------------------------------------------------//

procedure TBIFF8Stream.AddChunk(Tag,Size:word);
var
 header:cardinal;
begin
 header:=Tag+Size shl 16;
 WriteBuffer(header,SizeOf(header));
end;

procedure TBIFF8Stream.VarChunk(Tag:word; const Data; Size:word);
begin
 AddChunk(Tag,Size);
 WriteBuffer(Data,Size);
end;

//----------------------------------------------------------------------------//

constructor TWorkBook.Create;
begin
 inherited;
 fWorkSheets:=TStringList.Create;
 fSST:=TStringList.Create;
 fXFList:=TXFList.Create(Self);
 fFont:=TFont.Create;
 fFontList:=TFontList.Create(Self);
end;

destructor TWorkBook.Destroy;
begin
 fFont.Free;
 fFontList.Free;
 fXFList.Free;
 fSST.Free;
 fWorkSheets.Free;
 inherited;
end;

procedure TWorkbook.BeginDoc;
var
 BOF:TBIFF8_BOF;
 W1:TBIFF8_WINDOW1;
begin
 Position:=0;
 Size:=0;
// BOF
 BOF.Version     :=BIFF8_VERSION;
 BOF.DocType     :=BIFF8_WB_GLOBAL;
 BOF.BuildID     :=7422;
 BOF.BuildYear   :=1997;
 BOF.HistoryFlag :=16585;
 BOF.LowerVersion:=$0106;
 VarChunk(BIFF8_BOF,BOF,SizeOf(BOF));
// WINDOW1
 W1.HorzPos:=120;
 W1.VertPos:=120;
 W1.Width:=15180;
 W1.Height:=8835;
 W1.Flags:=8+16+32;
 W1.ActiveWS:=0;
 W1.ActiveTab:=0;
 W1.SelectTab:=0;
 W1.TabWidth:=600;
 VarChunk(BIFF8_WINDOW1,W1,SizeOf(W1));
// Font
// AddFontName('Arial',10,[]); // Headers ?
// AddFontName('Arial',10,[]); // Default
 fFont.Name:='Arial';
 fFont.Size:=10;
 fFont.Style:=[];
 fFontNumber:=fFontList.FindFont;
 fCurrentXF:=-1;
end;
(*
procedure TWorkbook.AddFontName(Name: string; Size: integer; Style: TFontStyles);
var
 BF:TBIFF8_FONT;
begin
 BF.Height:=20*Size;//200;
 BF.Flags :=byte(Style);
 BF.Color :=0;
 if fsBold in Style then BF.Weight:=700 else BF.Weight:=400;
 BF.Escape:=0;
 BF.Underline:=0;
 BF.Family:=0;
 BF.Charset:=0;
 BF.NameLen:=Length(Name);
 BF.NameFlag:=0;
 AddChunk(BIFF8_FONT,SizeOf(BF)+BF.NameLen);
 WriteBuffer(BF,SizeOf(BF));
 WriteBuffer(Name[1],BF.NameLen);
end;

procedure TWorkBook.AddFont(Font:TFont);
begin
 AddFontName(Font.Name,Font.Size,Font.Style);
end;
*)
procedure TWorkBook.SetFont(Value:TFont);
begin
 fFont.Assign(Value);
 fFontNumber:=-1;
 fCurrentXF:=-1;
end;

procedure TWorkBook.SetAlignment(Value:TCellAlignment);
begin
 fAlignment:=Value;
 fCurrentXF:=-1;
end;

function TWorkBook.GetBorder(Index:integer):TCellBorder;
begin
 Result:=TCellBorder((fBorders shr (4*Index)) and $F);
end;

procedure TWorkBook.SetBorder(Index:integer; Value:TCellBorder);
begin
 fBorders:=fBorders and (not ($F shl (4*Index))) or (ord(Value) shl (4*Index));
 fCurrentXF:=-1;
end;

function TWorkBook.GetCurrentXF:integer;
begin
 if fCurrentXF<0 then begin
  if fFontNumber<0 then fFontNumber:=fFontList.FindFont;
  fCurrentXF:=fXFList.FindFormat;
 end;
 Result:=fCurrentXF;
end;

function TWorkbook.AddWorkSheet(Name: string): TWorkSheet;
begin
 Result:=TWorkSheet.Create;
 Result.fWorkBook:=Self;
 fWorkSheets.AddObject(Name,Result);
end;

procedure TWorkbook.EndDoc;
var
 XF:TBIFF8_XF;
 ST:TBIFF8_STYLE;
 i,w:integer;
 SST:TBIFF8_SST;
 EXTSST:array of TBIFF8_EXTSST;
 e,x,start:integer;
 ofs:integer;
 BS:TBIFF8_BOUNDSHEET;
 s:string;
 ws:TWorkSheet;
begin
// FONT
 fFontList.Save;
// FORMAT
// XF
 fXFList.Save;
// STYLES
 ST.XF:=$8000; // XF 0
 ST.Style:=0;  // Normal
 ST.Level:=$FF;
 VarChunk(BIFF8_STYLE,ST,SizeOf(ST));
// SST
 if fSSTCount>0 then begin
  Setlength(EXTSST,(fSSTCount+7) div 8);
  e:=0;
  x:=0;
  SST.Total:=fSSTCount;
  SST.Count:=fSST.Count;
  w:=0;
  for i:=0 to fSST.Count-1 do begin
   inc(w,Length(fSST[i])+3);
  end;
  EXTSST[e].Count:=8;
  start:=Position;
  AddChunk(BIFF8_SST,SizeOf(SST)+w);
  WriteBuffer(SST,SizeOf(SST));
  for i:=0 to fSST.count-1 do begin
   s:=fSST[i];
   EXTSST[e].Strings[x].StreamOffset:=Position;
   EXTSST[e].Strings[x].RecordOffset:=Position-Start;
   inc(x);
   if x=8 then begin
    x:=0;
    inc(e);
    EXTSST[e].Count:=8;
   end;
   w:=Length(s);
   WriteBuffer(w,3);
   WriteBuffer(s[1],w);
  end;
  for i:=0 to Length(EXTSST)-2 do begin
   VarChunk(BIFF8_EXTSST,EXTSST[i],SizeOf(TBIFF8_EXTSST));
  end;
  VarChunk(BIFF8_EXTSST,EXTSST[e],2+x*8);
  EXTSST:=nil;
  fSST.Clear;
  fSSTCount:=0;
 end;

// BOUNDSHEETS
 ofs:=Position+fWorkSheets.Count*(2+2+SizeOf(BS))+2+2; // 2+2 for Tags + EOF
 for i:=0 to fWorkSheets.Count-1 do inc(ofs,Length(fWorkSheets[i]));
 for i:=0 to fWorkSheets.Count-1 do begin
  BS.Offset:=ofs;
  BS.Visibility:=0;
  BS.SheetType:=0;
  s:=fWorkSheets[i];
  BS.NameLen:=Length(s);
  BS.NameFlag:=0;
  AddChunk(BIFF8_BOUNDSHEET,SizeOf(BS)+BS.NameLen);
  WriteBuffer(BS,SizeOf(BS));
  WriteBuffer(s[1],BS.NameLen);
  inc(ofs,TWorkSheet(fWorkSheets.Objects[i]).Size);
 end;

// EOF
 AddChunk(BIFF8_EOF,0);

// WORKSHEETS
 for i:=0 to fWorkSheets.Count-1 do begin
  ws:=TWorkSheet(fWorkSheets.Objects[i]);
  ws.Close;
  Self.CopyFrom(ws,0);
  ws.Free;
 end;
 fWorkSheets.Clear;

 Position:=0;
end;

function TWorkbook.AddSST(s: string): integer;
begin
 Result:=fSST.IndexOf(s);
 if Result<0 then Result:=fSST.Add(s);
 inc(fSSTCount);
end;

//----------------------------------------------------------------------------//

constructor TWorkSheet.Create;
var
 BOF:TBIFF8_BOF;
begin
 BOF.Version     :=BIFF8_VERSION;
 BOF.DocType     :=BIFF8_WORKSHEET;
 BOF.BuildID     :=7422;
 BOF.BuildYear   :=1997;
 BOF.HistoryFlag :=16585;
 BOF.LowerVersion:=$0106;
 VarChunk(BIFF8_BOF,BOF,SizeOf(BOF));
end;

procedure TWorkSheet.Close;
begin
 AddChunk(BIFF8_EOF,0);
end;

function TWorkSheet.FloatToRK(Float:double; var RK:cardinal):boolean;
var
 DB:double;
 VV:array[0..1] of cardinal absolute DB;
begin
 DB:=Float;
 if (VV[0]=0) and ((VV[1] and 3)=0) then begin
  Result:=True;
  RK:=VV[1];
 end else begin
  DB:=DB*100;
  if (VV[0]=0) and ((VV[1] and 3)=0) then begin
   Result:=True;
   RK:=VV[1] or 1;
  end else begin
   Result:=False;
  end;
 end;
end;

function TWorkSheet.IntToRK(Int:integer; var RK:cardinal):boolean;
var
 top:integer;
begin
 top:=Int shr 30;
 if (top=0)or(top=3) then begin
  Result:=True;
  RK:=(Int shl 2) or 2;
 end else begin
  Result:=False;
 end;
end;

procedure TWorkSheet.AddRK(x, y: integer; Value: cardinal);
var
 RK:TBIFF8_RK;
begin
 RK.Row:=y;
 RK.Col:=x;
 RK.XF:=fWorkBook.Format;
 RK.Value:=Value;
 VarChunk(BIFF8_RK,RK,SizeOf(RK));
end;

procedure TWorkSheet.AddMULRK(x, y: integer; const RKS: array of cardinal);
var
 MRK:TBIFF8_MULRK;
 i:integer;
 w:word;
begin
 if Length(RKS)=1 then begin
  AddRK(x,y,RKS[0]);
 end else begin
  MRK.Row:=y;
  MRK.Col:=x;
  AddChunk(BIFF8_MULRK,SizeOf(MRK)+6*Length(RKS)+2);
  WriteBuffer(MRK,SizeOf(MRK));
  w:=fWorkBook.Format;
  for i:=Low(RKS) to High(RKS) do begin
   WriteBuffer(w,2); // MULRK allow multiple XF...
   WriteBuffer(RKS[i],4);
  end;
  w:=x+Length(RKS)-1;
  WriteBuffer(w,2);
 end;
end;

procedure TWorkSheet.AddLabel(x, y: integer; Value: string);
var
 LBL:TBIFF8_LABELSST;
begin
 LBL.Row:=y;
 LBL.Col:=x;
 LBL.XF :=fWorkBook.Format;
 LBL.SST:=fWorkBook.AddSST(Value);
 VarChunk(BIFF8_LABELSST,LBL,SizeOf(LBL));
end;

procedure TWorkSheet.AddNumber(x, y: integer; Value: double);
var
 NB:TBIFF8_NUMBER;
begin
 NB.Row:=y;
 NB.Col:=x;
 NB.XF:=fWorkBook.Format;
 NB.Value:=Value;
 VarChunk(BIFF8_NUMBER,NB,SizeOf(NB));
end;

procedure TWorkSheet.AddFloat(x, y: integer; Value: double);
var
 RK:cardinal;
begin
 if FloatToRK(Value,RK) then AddRK(x,y,RK) else AddNumber(x,y,Value);
end;

procedure TWorkSheet.AddFloats(x, y: integer; Values:array of double);
var
 rk :integer;
 rks:array[0..255] of cardinal;
 rkx:integer;
   i:integer;
   v:double;
   b:boolean;
begin
 rk:=0;
 rkx:=x;
 for i:=Low(Values) to High(Values) do begin
  v:=Values[i];
  b:=FloatToRK(v,rks[rk]);
  if b then begin
   if rk=0 then rkx:=x+i;
   inc(rk);
  end;
  if (rk=256) or ((b=False)and(rk>0)) then begin
   AddMULRK(rkx,y,slice(rks,rk));
   rk:=0;
  end;
  if b=False then AddNumber(x+i,y,v);
 end;
 if rk>0 then AddMULRK(rkx,y,slice(rks,rk));
end;

procedure TWorkSheet.AddInteger(x, y, Value: integer);
var
 RK:cardinal;
begin
 If IntToRK(Value,RK) then AddRK(x,y,RK) else AddNumber(x,y,Value);
end;

procedure TWorkSheet.AddIntegers(x, y:integer; Values:array of integer);
var
 rk :integer;
 rks:array[0..255] of cardinal;
 rkx:integer;
   i:integer;
   v:integer;
   b:boolean;
begin
 rk:=0;
 for i:=Low(Values) to High(Values) do begin
  v:=Values[i];
  b:=IntToRK(v,rks[rk]);
  if b then begin
   if rk=0 then rkx:=x+i;
   inc(rk);
  end;
  if (rk=256) or ((b=False)and(rk>0)) then begin
   AddMULRK(rkx,y,slice(rks,rk));
   rk:=0;
  end;
  if b=False then AddNumber(x+i,y,v);
 end;
 if rk>0 then AddMULRK(rkx,y,slice(rks,rk));
end;

//----------------------------------------------------------------------------//

constructor TFontList.Create(WorkBook:TWorkBook);
var
 i:integer;
 f:TFontEntry;
begin
 inherited Create;
 fWorkBook:=WorkBook;
 for i:=0 to 15 do begin
  f:=TFontEntry.Create;
  f.Name:='Arial';
  f.Size:=10+i;
  Add(f);
 end;{
 f:=TFontEntry.Create;
 f.Name:='Arial';
 f.Size:=10;
 f.Style:=[fsStrikeOut];
 Add(f); }
end;

function TFontList.FindFont:integer;
var
 f:TFontEntry;
 i:integer;
begin
 for i:=1 to Count-1 do begin
  f:=List[i];
  if (f.Name=fWorkBook.Font.Name)
  and(f.Size=fWorkBook.Font.Size)
  and(f.Style=fWorkBook.Font.Style) then begin
   if i<4 then Result:=i else Result:=i+1;
   Result:=i;
   exit;
  end;
 end;
 f:=TFontEntry.Create;
 f.Name:=fWorkBook.Font.Name;
 f.Size:=fWorkBook.Font.Size;
 f.Style:=fWorkBook.Font.Style;
 Result:=Add(f);
 if Result>=4 then inc(i);
end;

procedure TFontList.Save;
var
  i:integer;
  f:TFontEntry;
 FN:TBIFF8_FONT;
begin
 FillChar(FN,SizeOf(FN),0);
 for i:=0 to Count-1 do begin
  f:=List[i];
  FN.Height:=20*f.Size;
  FN.Flags :=Byte(f.Style);
  if fsBold in f.Style then FN.Weight:=700 else FN.Weight:=400;
  FN.NameLen:=Length(f.Name);
  fWorkBook.AddChunk(BIFF8_FONT,SizeOf(FN)+FN.NameLen);
  fWorkBook.WriteBuffer(FN,SizeOf(FN));
  fWorkBook.WriteBuffer(f.Name[1],FN.NameLen);
 end;
 Clear;
end;

//----------------------------------------------------------------------------//

constructor TXFList.Create(WorkBook:TWorkBook);
var
 i:integer;
 f:TExtendFormat;
begin
 inherited Create;
 fWorkBook:=WorkBook;
 for i:=0 to 14 do begin
  f:=TExtendFormat.Create;
  f.Used:=True;
  f.Font:=i;
  f.Parent:=$FFF5;
  f.Flags:=$F8;
  Add(f);
 end;
 for i:=15 to 15+15 do begin
  f:=TExtendFormat.Create;
  f.Used:=True;
  f.Font:=i-15;
  f.Parent:=i-15;
  f.Flags:=$F8;
  Add(f);
 end;
(*
 f:=TExtendFormat.Create;
 f.Used:=True;
 f.Parent:=$FFF5;
 f.Flags:=$F8;
 Add(f);
 for i:=2 to 14 do begin
  f:=TExtendFormat.Create;
  f.Used:=True;
  f.Parent:=$FFF5;
  f.Font:=1;
  f.Parent:=0;
  Add(f);
 end;
 f:=TExtendFormat.Create;
 f.Used:=True;
 f.Font:=1;
 f.Parent:=0;
 f.Flags:=0;
 Add(f);
*)
end;

function TXFList.FindFormat:integer;
var
 i,e1,e2,p:integer;
 f:TExtendFormat;
begin
 e1:=-1;
 e2:=-1;
 p:=-1;
 for i:=Count-1 downto 0 do begin
  f:=List[i];
  if f.Used=False then begin
   e2:=e1;
   e1:=i;
  end else begin
   if (f.Font=fWorkBook.fFontNumber)
   and(f.Alignment=fWorkBook.fAlignment)
   and(f.Borders=fWorkBook.fBorders)then begin
    if f.Parent=$FFF5 then
     p:=i
    else begin
     Result:=i;
     exit;
    end;
   end;
  end;
 end;
 // Style XF
 if p<0 then begin
  if e1<0 then begin
   f:=TExtendFormat.Create;
   f.Used:=True;
   p:=Add(f);
  end else begin
   f:=List[e1];
   p:=e1;
   e1:=e2;
  end;
  f.Font:=fWorkBook.fFontNumber;
  f.Parent:=$FFF5;
  f.Alignment:=fWorkBook.fAlignment;
  f.Flags:=$F4;
  f.Borders:=fWorkBook.fBorders;
 end;
 // Cell XF
 if e1<0 then begin
  f:=TExtendFormat.Create;
  f.Used:=True;
  Result:=Add(f);
 end else begin
  f:=List[e1];
  f.Used:=True;
  Result:=e1;
 end;
 f.Parent:=p;
 f.Font:=fWorkBook.fFontNumber;
 f.Alignment:=fWorkBook.fAlignment;
 f.Flags:=$F4;
 f.Borders:=fWorkBook.fBorders;
end;

procedure TXFList.Save;
var
 i:integer;
 f:TExtendFormat;
 XF:TBIFF8_XF;
begin
 FillChar(XF,SizeOf(XF),0);
 XF.Attr3:=$20C0;
 for i:=0 to Count-1 do begin
  f:=List[i];
  XF.Font:=f.Font;
  XF.Parent:=f.Parent;
  XF.Align:=byte(f.Alignment) or $20 { Bottom } ;
  XF.Flags:=f.Flags;
  XF.Attr1 :=f.Borders;
  fWorkBook.VarChunk(BIFF8_XF,XF,SizeOf(XF));
 end;
 Clear;
end;

end.
